############### Begin Standard Header - Do not add comments here ###############
#
# File:     generic.py
# Version:  1.2
# Modified: 01/22/09
#
# Licensed Materials - Property of IBM
#
# 5724-N55
#
# (C) COPYRIGHT IBM CORP. 2009.  All Rights Reserved.
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
#
############################# End Standard Header ##############################
import sys
import java

from java.lang import System
from java.lang import Object

coll_home = System.getProperty("com.collation.home")
jython_home = coll_home + "/external/jython-2.1"

System.setProperty("jython.home",jython_home)
System.setProperty("python.home",jython_home)

sys.path.append(jython_home + "/Lib")
sys.path.append(coll_home + "/lib/sensor-tools")
sys.prefix = jython_home + "/Lib"

import traceback
import string
import re
import jarray
import sensorhelper
import zlib
from java.util import Properties
from java.io import FileInputStream
from java.lang import Class
from com.collation.platform.logger import LogFactory

#----------------------------------------------------------------------------------
# Define some CONSTANTS, print some info
#-----------------------------------------------------------------------------------

prop_prefix = "discover."

#----------------------------------------------------------------------------------
# Define the Functions
#-----------------------------------------------------------------------------------

########################
# LogError      Error logger - raises exception and ends script
########################
def LogError(msg):
        '''
        Print Error Message using Error Logger with traceback information
        This also raises a RuntimeError and ends processing of the script
        '''
        error_text = template_name + ": " + msg
        log.error(error_text)
        (ErrorType, ErrorValue, ErrorTB) = sys.exc_info()
        traceback.print_exc(ErrorTB)
        result.warning(error_text)
        raise RuntimeError, error_text

########################
# LogWarning      Warning logger
########################
def LogWarning(msg):
        '''
        Print Warning Message using Error Logger with traceback information
        '''
        warning_text = template_name + ": " + msg
        log.warning(warning_text)
        result.warning(warning_text)

########################
# LogDebug      Print routine for normalized messages in log
########################
def LogDebug(msg):
        '''
        Print Debug Message using debug logger (from sensorhelper)
        '''
        log.debug(template_name + ": " + msg)

########################
# LogInfo Print routine for normalized messages in log
########################
def LogInfo(msg):
        '''
        Print INFO level Message using info logger (from sensorhelper)

        '''
        log.info(template_name + ": " + msg)

def get_os_type(appserver):
        '''
        Return OS we are on --> UNIX or WINDOWS
        '''
        os_name = appserver.getHost().getOSRunning().getOSName()
        if re.search("AIX|HP-UX|Linux|SunOS",os_name,re.I):
                os_type = "UNIX"
        elif re.search("Windows",os_name,re.I):
                os_type = "WINDOWS"
        else:
                os_type = "UNKNOWN"

        LogInfo("OS_NAME: Type --> " + str(os_type) + " running on " + str(os_name))
        return os_type


########################
# run_customizations - can be use to add additional functions as needed
########################
def run_customizations():
        '''
        example:  if template_name == "MY_TEMPLATE":
                          LogInfo("Running customization code for MY_TEMPLATE.")
                          <custom code here>
        '''
        LogInfo("Running customizations")
        #
        # put customizations here
        #
        LogInfo("Completed running customizations.")


def replace_envVariables(string):
        '''
        Return string with environment variable replaced
        '''
        original_string = string
        LogDebug("Replacing environment values in (" + str(string) + ").")
        if os_type == "WINDOWS":
                exp="%([a-zA-Z_0-9]+)%"
                beginChar = "%"
                endChar = "%"
        else:
                exp="\$({?[a-zA-Z_0-9]+}?)"
                beginChar = "\$"
                endChar = ""

        LogDebug("Looking for expression (" + exp + ").")
        # finditer doesn't seem to work with our Jython version
        matches = re.findall(exp, string)
        for match in matches:
                replacementText = beginChar + match + endChar
                if match.startswith("{"):
                        if match.endswith("}"):
                                match = re.sub("^{","",match)
                                match = re.sub("}$","",match)
                        else:
                                LogDebug("Match found with left parenthesis but no right parenthesis.")
                LogDebug("Looking for environment variable (" + str(match) + ").")
                if envVariables.has_key(match):
                        replacementValue = envVariables[match]
                        LogDebug("Replacing (" + str(replacementText) + ") with (" + str(replacementValue) + ") in starting string (" + str(string) + ").")
                        replacementValue = re.sub(r"\\",r"\\\\",replacementValue)
                        LogDebug("Replacing (" + str(replacementText) + ") with (" + str(replacementValue) + ") in starting string (" + str(string) + ").")
                        string = re.sub(replacementText,replacementValue,string)
                        LogDebug("Resulting string (" + str(string) + ").")
                else:
                        LogDebug("No enviroment variable (" + str(replacementText) + ") found.  Ignoring...")
        LogDebug("Original string (" + str(original_string) + ").")
        LogDebug("Final string (" + str(string) + ").")
        return string


def parseEnv(environment):
        '''
        Set up environments from envirionment string
        '''
        LogDebug("Gathering environment variables from env (" + environment + ").")

        if os_type == "WINDOWS":
                variables = environment.splitlines()
        else:
                variables = environment.split(" ")

        for variable in variables:
                LogDebug("     Variable: " + variable)
                tokens = variable.split("=")
                if len(tokens) != 2:
                        LogDebug("Ignoring variable (" + variable + ") with unexpected format.")
                else:
                        LogDebug("     Token 1: " + tokens[0])
                        LogDebug("     Token 2: " + tokens[1])
                        envVariables[tokens[0]] = tokens[1]


def get_process_directory():
        '''
        Return process directory from the command line
        '''
        LogInfo("Getting process directory for commandline: " + str(original_commandline))
        process_dir = ""

        regexp = "^\"(.*?)\""
        m1  = re.match(regexp,original_commandline)
        if m1 == None:
                i=0
                tokens = original_commandline.split(" ")
                for token in tokens:
                         LogDebug("token="+token)
                         i=i+1
                         if i==1:
                                  program=token
        else:
                program=m1.group(1)
        LogDebug("program: "+program)

        if os_type == "WINDOWS":
                exp="\/"
                m = re.search(exp,program)
                if m == None:
                         LogDebug("Parsing windows command (" + program + ") using backward slash as path delimiter.")
                         expression="(^.*)\\\.?"
                else:
                         LogDebug("Parsing windows command (" + program + ") using forward slash as path delimiter.")
                         expression="(^.*)/.?"
        else:
                expression="(^.*)/.?"
        match = re.match(expression,program)
        if match == None:
                #didnt match
                LogDebug("Unable to parse command line (" + original_commandline + ") using program (" + program + ") and expression (" + expression + ").")
                return process_dir

        temp_process_dir = match.group(1)
        process_dir=re.sub(r"^\"",r"",temp_process_dir)
        if os_type == "WINDOWS":
                LogDebug("Updating process directory path for windows: " + process_dir)
                process_dir=re.sub(r"\\",r"\\\\",process_dir)

        # put process_dir in the variable list
        vars["processdir"]=process_dir

        LogDebug("Returning process directory: " + process_dir)
        return process_dir


def create_app_config(label,command,ignore):
        '''
        Run Command and place output in AppConfig object of AppServer
        '''
        LogDebug("create_app_config called with label (" + str(label) + "), command (" + str(command) + ") , and ignore (" + str(ignore) + ").")
        config = run_command(command, ignore)
        if config != "":
                uri = "config://" + bind_address + "/" + template_name + "::" + label
                configobject = sensorhelper.newModelObject("com.collation.platform.model.topology.app.AppConfig")
                configobject.setLabel(str(label))
                LogDebug("Created AppConfig 'configobject'")
                configcontent = sensorhelper.newModelObject("com.collation.platform.model.topology.sys.DataFile")
                config_crc = zlib.crc32(config)
                config_size = len(config)
                LogDebug("CONFIG: " + str(config_size) + " CRC: " + str(config_crc))
                configcontent.setContent(config)
                LogDebug("Completed 'configcontent.setContent'")
                configcontent.setChecksum(str(config_crc))
                configcontent.setSize(config_size)
                configcontent.setURI(uri)
                LogDebug("Completed 'configcontent.setURI("+ uri + ")'")
                configcontent.setFixedPath(label)
                configobject.setContent(configcontent)
                LogDebug("Completed 'configobject.setContent'")
                configobject.setParent(app_server)
                result.setConfigContent(configobject)
        else:
                LogDebug("No output from command (" + str(command) + ").  Skipping setting creating config content for label (" + str(label) + ").")


def create_default_app_config(data):
        '''
        Create the default app config file for the template
        '''
        LogDebug("create_default_app_config called with data (" + str(data) + ").")
        label = template_name + ".config"
        uri = "config://" + bind_address + "/" + template_name + "/" + label
        configobject = sensorhelper.newModelObject("com.collation.platform.model.topology.app.AppConfig")
        configobject.setLabel(str(label))
        LogDebug("Created 'configcontent'")
        configcontent = sensorhelper.newModelObject("com.collation.platform.model.topology.sys.DataFile")
        LogDebug("'configcontent.setContent'")
        config_crc = zlib.crc32(data)
        config_size = len(data)
        LogDebug("CONFIG: " + str(config_size) + " CRC: " + str(config_crc))
        configcontent.setContent(data)
        configcontent.setChecksum(str(config_crc))
        configcontent.setSize(config_size)

        LogDebug("'configcontent.setURI("+ uri + ")'")
        configcontent.setURI(uri)
        configcontent.setFixedPath(label)
        LogDebug("'configobject.setContent'")
        configobject.setContent(configcontent)
        configobject.setParent(app_server)
        result.setConfigContent(configobject)


def run_command(command, ignoreError):
        '''
        Run the command only
        '''
        stripped_result = ""
        LogDebug("run_command called with command (" + str(command) + ") and ignoreError (" + str(ignoreError) + ").")
        try:
                result = sensorhelper.executeCommand(command,os_handle)
                stripped_result = str(result).strip()
                LogDebug("run_command completed command (" + str(command) + ") with result (" + result + ").")
        except:
                LogDebug("Exception in run_command.")
                if ignoreError:
                        LogDebug("Ignoring error as directed.")
                        return ""
                else:
                        LogWarning("run_command failed on  command (" + str(command) + ").")
        return stripped_result


def set_attr_with_command(attr, command, ignoreError):
        '''
        Run Command and set attr value in the app_server to the command output
        '''
        LogDebug("set_attr_with_command called with attr (" + str(attr) + "), command (" + str(command) + ") and ignoreError (" + str(ignoreError) + ").")
        output = run_command(command, ignoreError)
        LogDebug(attr + " command result: " + output + ".")
        if output != "":
                methodName="set"+attr.lower()
                LogDebug("method name: " + methodName)
                for method in app_server.getClass().getMethods():
                        as_methodName = method.getName().lower()
                        LogDebug("app_server method name: " + as_methodName)
                        if as_methodName == methodName:
                                LogDebug("Matched methodName: " + method.getName())
                                if len(method.getParameterTypes()) != 1:
                                        continue
                                params = [str(output)]
                                LogDebug("Invoking: " + method.getName())
                                method.invoke(app_server, params)
                                return
                LogWarning("No set method found for attribute: " + attr)
        else:
                LogDebug("No output from command (" + str(command) + ").  Skipping setting attribute (" + str(attr) + ").")



def set_attr_to_value(attr, value):
        '''
        set attr value in the app_server to the value
        '''
        LogDebug("set_attr_to_value called with attr (" + str(attr) + ") and value (" + str(value) + ").")

        for var, replacementValue in vars.items():
                replacementText = "&&" + var + "&&"
                value = re.sub(replacementText,replacementValue,value)
                LogDebug("Value after replacement of (" + replacementText + ") with (" + replacementValue + "): " + value)

        methodName="set"+attr.lower()
        LogDebug("method name: " + methodName)
        for method in app_server.getClass().getMethods():
                as_methodName = method.getName().lower()
                LogDebug("app_server method name: " + as_methodName)
                if as_methodName == methodName:
                        LogDebug("Matched methodName: " + method.getName())
                        if len(method.getParameterTypes()) != 1:
                             continue

                        params = [str(value)]
                        LogDebug("Invoking: " + method.getName())
                        method.invoke(app_server, params)
                        return
        LogWarning("No set method found for attribute: " + attr)


def process_props(props):
        '''
        initializes configFiles, configCommands, attrcommands, attrvalues, commands, and attrConfigCommands
        '''
        enumer = props.propertyNames()
        while enumer.hasMoreElements() == 1:
                property = enumer.nextElement()
                LogDebug("Processing Property: " + str(property))

                if os_type == "WINDOWS":
                        attrvalue_expression=prop_prefix + "windows\.attrvalue\.(.*)"
                        cfgfile_expression=prop_prefix + "windows\.configfile(-ignore)*\.+"
                        cfgcmd_expression=prop_prefix + "windows\.configcmd(-ignore)*\.+"
                        attrcmd_expression=prop_prefix + "windows\.attrcmd(-ignore)*\.(.*)"
                        cmd_expression=prop_prefix + "windows\.cmd(-ignore)*\.+"
                        attrcfgcmd_expression=prop_prefix + "windows\.attrconfigcmd(-ignore)*\.(.*)"
                        var_expression=prop_prefix + "windows\.var\.(.*)"
                        varcmd_expression=prop_prefix + "windows\.varcmd(-ignore)*\.(.*)"
                        varpath_expression=prop_prefix + "windows\.varpath\.(.*)"
                        varpathcmd_expression=prop_prefix + "windows\.varpathcmd(-ignore)*\.(.*)"
                        varselect_expression=prop_prefix + "windows\.varselect\.(.*)"
                        varregex_expression=prop_prefix + "windows\.varregex\.(.*)"
                        regcfgfile_expression=prop_prefix + "windows\.regconfigfile(-ignore)*\.(.*)"
                else:
                        attrvalue_expression=prop_prefix + "unix\.attrvalue\.(.*)"
                        cfgfile_expression=prop_prefix + "unix\.configfile(-ignore)*\.+"
                        attrcmd_expression=prop_prefix + "unix\.attrcmd(-ignore)*\.(.*)"
                        cfgcmd_expression=prop_prefix + "unix\.configcmd(-ignore)*\.+"
                        cmd_expression=prop_prefix + "unix\.command(-ignore)*\.+"
                        attrcfgcmd_expression=prop_prefix + "unix\.attrconfigcmd(-ignore)*\.(.*)"
                        var_expression=prop_prefix + "unix\.var\.(.*)"
                        varcmd_expression=prop_prefix + "unix\.varcmd(-ignore)*\.(.*)"
                        varpath_expression=prop_prefix + "unix\.varpath\.(.*)"
                        varpathcmd_expression=prop_prefix + "unix\.varpathcmd(-ignore)*\.(.*)"
                        varselect_expression=prop_prefix + "unix\.varselect\.(.*)"
                        varregex_expression=prop_prefix + "unix\.varregex\.(.*)"
                        regcfgfile_expression="^$"

                # is it a config file?
                cfgfile_match = re.match(cfgfile_expression,property)
                if cfgfile_match == None:

                        # is it a configcommand?
                        cfgcmd_match = re.match(cfgcmd_expression,property)
                        if cfgcmd_match == None:

                                # is it an attrcommand?
                                attrcmd_match = re.match(attrcmd_expression,property)
                                if attrcmd_match == None:

                                        # is it an attrvalue?
                                        attrvalue_match = re.match(attrvalue_expression,property)
                                        if attrvalue_match == None:

                                                # is it a command?
                                                cmd_match = re.match(cmd_expression,property)
                                                if cmd_match == None:

                                                        # is it an attrConfigCommand?
                                                        attrcfgcmd_match = re.match(attrcfgcmd_expression,property)
                                                        if attrcfgcmd_match == None:

                                                                # is it a var?
                                                                var_match = re.match(var_expression,property)
                                                                if var_match == None:

                                                                        # is it a varcmd?
                                                                        varcmd_match = re.match(varcmd_expression,property)
                                                                        if varcmd_match == None:

                                                                                # is it a varpath?
                                                                                varpath_match = re.match(varpath_expression,property)
                                                                                if varpath_match == None:

                                                                                        # is it a varpathcmd?
                                                                                        varpathcmd_match = re.match(varpathcmd_expression,property)
                                                                                        if varpathcmd_match == None:

                                                                                                # is it a varselect?
                                                                                                varselect_match = re.match(varselect_expression,property)
                                                                                                if varselect_match == None:

                                                                                                        # is it a varregex?
                                                                                                        varregex_match = re.match(varregex_expression,property)
                                                                                                        if varregex_match == None:

                                                                                                                # is it a regcfgfile?
                                                                                                                regcfgfile_match = re.match(regcfgfile_expression,property)
                                                                                                                if regcfgfile_match == None:

                                                                                                                        #didn't match anything
                                                                                                                        LogDebug("Ignoring unrecognized property (" + str(property) + ").")
                                                                                                                        continue

                                                                                                                # it is a regcfgfile
                                                                                                                LogDebug("Registry Config File property found: " + str(property))
                                                                                                                regconfigfile = props.getProperty(property)
                                                                                                                ignore = 0
                                                                                                                if regcfgfile_match.group(1) == "-ignore":
                                                                                                                        ignore = 1
                                                                                                                LogDebug("Registry key: " + regconfigfile)
                                                                                                                LogDebug("ignoreError: " + str(ignore))
                                                                                                                regconfigfile = replace_envVariables(regconfigfile)
                                                                                                                LogDebug("Registry key with expanded environment variables: " + str(regconfigfile))
                                                                                                                regConfigFiles[regconfigfile] = ignore
                                                                                                                continue

                                                                                                        # it is a varregex
                                                                                                        LogDebug("Variable Regex property found: " + property)
                                                                                                        var = varregex_match.group(1)
                                                                                                        LogDebug("Variable: " + var)
                                                                                                        varregex=props.getProperty(property)
                                                                                                        LogDebug("Regex: " + varregex)
                                                                                                        varregexs[var]=varregex
                                                                                                        continue

                                                                                                # it is a varselect
                                                                                                LogDebug("Variable Select property found: " + property)
                                                                                                varselect = varselect_match.group(1)
                                                                                                LogDebug("Selection Variable: " + varselect)
                                                                                                varselectionlist=props.getProperty(property)
                                                                                                LogDebug("Variable Selection List: " + varselectionlist)
                                                                                                varselects[varselect]=varselectionlist
                                                                                                continue

                                                                                        # it is a varpathcmd
                                                                                        LogDebug("Variable Path Command property found: " + property)
                                                                                        ignore = 0
                                                                                        if varpathcmd_match.group(1) == "-ignore":
                                                                                                ignore = 1
                                                                                        varpathcmd = varpathcmd_match.group(2)
                                                                                        LogDebug("Replacement Path Variable: " + varpathcmd)
                                                                                        LogDebug("ignoreError: " + str(ignore))
                                                                                        replacementpathcmd=props.getProperty(property)
                                                                                        LogDebug("Replacement Path Command: " + replacementpathcmd)
                                                                                        replacementpathcmd = replace_envVariables(replacementpathcmd)
                                                                                        LogDebug("Replacement Path Command with expanded environment variables: " + str(replacementpathcmd))
                                                                                        if isFullyExpanded(replacementpathcmd):
                                                                                                replacementpathvalue=run_command(replacementpathcmd, ignore)
                                                                                                if os_type == "WINDOWS":
                                                                                                        LogDebug("Updating replacement path for windows: " + str(replacementpathvalue))
                                                                                                        replacementpathvalue=re.sub(r"\\",r"\\\\",replacementpathvalue)
                                                                                                LogDebug("Replacement Path Value: " + str(replacementpathvalue))
                                                                                                vars[varpathcmd]=replacementpathvalue
                                                                                        else:
                                                                                                LogDebug("Replacement Path Variable is not fully expanded.  Saving to run after expansion.")
                                                                                                unexpandedVarPathCmds[varpathcmd]=[replacementpathcmd,ignore]
                                                                                        continue

                                                                                # it is a varpath
                                                                                LogDebug("Variable Path property found: " + property)
                                                                                varpath = varpath_match.group(1)
                                                                                LogDebug("Replacement Path Variable: " + varpath)
                                                                                replacementpathvalue=props.getProperty(property)
                                                                                LogDebug("Replacement Path Value: " + replacementpathvalue)
                                                                                replacementpathvalue = replace_envVariables(replacementpathvalue)
                                                                                LogDebug("Replacement Path Value with expanded environment variables: " + str(replacementpathvalue))
                                                                                if os_type == "WINDOWS":
                                                                                        LogDebug("Updating initial replacement path variable for windows: " + str(replacementpathvalue))
                                                                                        replacementpathvalue=re.sub(r"\\",r"\\\\",replacementpathvalue)
                                                                                if isFullyExpanded(replacementpathvalue):
                                                                                        LogDebug("Replacement Path Value: " + replacementpathvalue)
                                                                                        vars[varpath]=replacementpathvalue
                                                                                else:
                                                                                        LogDebug("Replacement Path Value is not fully expanded.  Saving to expand later.")
                                                                                        unexpandedVarPaths[varpath]=replacementpathvalue
                                                                                continue

                                                                        # it is a varcmd
                                                                        LogDebug("Variable Command property found: " + property)
                                                                        ignore = 0
                                                                        if varcmd_match.group(1) == "-ignore":
                                                                                ignore = 1
                                                                        varcmd = varcmd_match.group(2)
                                                                        LogDebug("Command Replacement Variable: " + varcmd)
                                                                        LogDebug("ignoreError: " + str(ignore))
                                                                        replacementcmd=props.getProperty(property)
                                                                        LogDebug("Replacement Command: " + replacementcmd)
                                                                        replacementcmd = replace_envVariables(replacementcmd)
                                                                        LogDebug("Replacement Command with expanded environment variables: " + str(replacementcmd))
                                                                        if isFullyExpanded(replacementcmd):
                                                                                replacementvalue=run_command(replacementcmd, ignore)
                                                                                LogDebug("Command Replacement Variable Value: " + str(replacementvalue))
                                                                                vars[varcmd]=replacementvalue
                                                                        else:
                                                                                LogDebug("Command Replacement Variable is not fully expanded.  Saving to run after expansion.")
                                                                                unexpandedVarCmds[varcmd]=[replacementcmd,ignore]
                                                                        continue

                                                                # it is a var
                                                                LogDebug("Variable property found: " + property)
                                                                var = var_match.group(1)
                                                                LogDebug("Replacement Variable: " + var)
                                                                replacementvalue=props.getProperty(property)
                                                                LogDebug("Replacement Variable Value: " + replacementvalue)
                                                                replacementvalue = replace_envVariables(replacementvalue)
                                                                LogDebug("Replacement Variable Value with expanded environment variables: " + str(replacementvalue))
                                                                if isFullyExpanded(replacementvalue):
                                                                        LogDebug("Replacement Variable Value: " + str(replacementvalue))
                                                                        vars[var]=replacementvalue
                                                                else:
                                                                        LogDebug("Replacement Variable is not fully expanded.  Saving to expand later.")
                                                                        unexpandedVars[var]=replacementvalue
                                                                continue

                                                        # it is an attrConfigCommand
                                                        LogDebug("AttrConfigCommand property found: " + str(property))
                                                        ignore = 0
                                                        if attrcfgcmd_match.group(1) == "-ignore":
                                                                ignore = 1
                                                        key = attrcfgcmd_match.group(2)
                                                        LogDebug("Key: " + key)
                                                        LogDebug("ignoreError: " + str(ignore))
                                                        keyvalue=props.getProperty(property)
                                                        LogDebug("Value: " + keyvalue)
                                                        keyvalue = replace_envVariables(keyvalue)
                                                        LogDebug("Value with expanded environment variables: " + str(keyvalue))
                                                        attrConfigCommands[key]=[keyvalue,ignore]
                                                        continue

                                                # it is a command
                                                LogDebug("Command property found: " + str(property))
                                                cmd = props.getProperty(property)
                                                ignore = 0
                                                if cmd_match.group(1) == "-ignore":
                                                        ignore = 1
                                                LogDebug("Command: " + cmd)
                                                LogDebug("ignoreError: " + str(ignore))
                                                cmd = replace_envVariables(cmd)
                                                LogDebug("Command with expanded environment variables: " + str(cmd))
                                                commands[cmd] = ignore
                                                continue

                                        # it is an attrvalue
                                        LogDebug("AttrValue property found: " + str(property))
                                        attr = attrvalue_match.group(1)
                                        LogDebug("Attr: " + str(attr))
                                        attrvalue=props.getProperty(property)
                                        LogDebug("Value: " + str(attrvalue))
                                        attrvalue = replace_envVariables(attrvalue)
                                        LogDebug("Value with expanded environment variables: " + str(attrvalue))
                                        attrvalues[attr]=attrvalue
                                        continue

                                # it is an attrcommand
                                LogDebug("AttrCommand property found: " + str(property) + " -- attrcmd_match: " + str(attrcmd_match))
                                ignore = 0
                                if attrcmd_match.group(1) == "-ignore":
                                        ignore = 1
                                attr = attrcmd_match.group(2)
                                LogDebug("Attr: " + str(attr))
                                LogDebug("ignoreError: " + str(ignore))
                                attrcmd=props.getProperty(property)
                                LogDebug("Command: " + str(attrcmd))
                                attrcmd = replace_envVariables(attrcmd)
                                LogDebug("Command with expanded environment variables: " + str(attrcmd))
                                attrCommands[attr]=[attrcmd, ignore]
                                continue

                        # it is a configcommand?
                        LogDebug("ConfigCommand property found: " + str(property))
                        cfgcmd=props.getProperty(property)
                        ignore = 0
                        if cfgcmd_match.group(1) == "-ignore":
                                ignore = 1
                        LogDebug("ConfigCommand: " + str(cfgcmd))
                        LogDebug("ignoreError: " + str(ignore))
                        cfgcmd = replace_envVariables(cfgcmd)
                        LogDebug("ConfigCommand with expanded environment variables: " + str(cfgcmd))
                        configCommands[cfgcmd] = ignore
                        continue

                # it is a config file
                LogDebug("Config file property found: " + str(property))
                filename=props.getProperty(property)
                ignore = 0
                if cfgfile_match.group(1) == "-ignore":
                        ignore = 1
                LogDebug("Config file: " + str(filename))
                LogDebug("ignoreError: " + str(ignore))
                filename = replace_envVariables(filename)
                LogDebug("Config file with expanded environment variables: " + str(filename))
                configFiles[filename] = ignore


def define_regex_matches():
        '''
        set variables for any regex groupings
        '''
        for var, regex in varregexs.items():
                # if variable has been expanded, process it.  Otherwise ignore it.
                if vars.has_key(var):
                        # check that it matches
                        value = vars[var]
                        LogDebug("Processing regex (" + str(regex) + ") on variable (" + str(var) + ") with value (" + value + ").")
                        regex_match = re.match( str(regex),value)
                        if regex_match == None:
                                LogWarning("Regex (" + str(regex) + ") does not match variable (" + str(var) + ") value (" + value + ").")
                                continue
                        LogDebug("Regex (" + regex + ") matches variable (" + var + ") value (" + value + ").")
                        try:
                                group1 = regex_match.group(1)
                                LogDebug("Group 1: " + group1)
                                vars[var+"1"] = group1
                                group2 = regex_match.group(2)
                                LogDebug("Group 2: " + group2)
                                vars[var+"2"] = group2
                                group3 = regex_match.group(3)
                                LogDebug("Group 3: " + group3)
                                vars[var+"3"] = group3
                                group4 = regex_match.group(4)
                                LogDebug("Group 4: " + group4)
                                vars[var+"4"] = group4
                                group5 = regex_match.group(5)
                                LogDebug("Group 5: " + group5)
                                vars[var+"5"] = group5
                                group6 = regex_match.group(6)
                                LogDebug("Group 6: " + group6)
                                vars[var+"6"] = group6
                                group7 = regex_match.group(7)
                                LogDebug("Group 7: " + group7)
                                vars[var+"7"] = group7
                                group8 = regex_match.group(8)
                                LogDebug("Group 8: " + group8)
                                vars[var+"8"] = group8
                                group9 = regex_match.group(9)
                                LogDebug("Group 9: " + group9)
                                vars[var+"9"] = group9
                                group10 = regex_match.group(10)
                                LogDebug("Group 10: " + group10)
                                vars[var+"10"] = group10
                        except:
                                LogDebug("End of matched groups.")
                                continue
                else:
                        LogDebug("No variable (" + str(var) + ") found.  Ignoring...")
                        continue

def expand_variables(last_varKeys):
        '''
        expand any variable commands containing other variables
        '''
        if not unexpandedVarPathCmds and not unexpandedVarCmds and not unexpandedVars and not unexpandedVarPaths:
                # all done
                LogDebug("Finished expanding all unexpanded variables.")
                return

        for varA in last_varKeys:
             LogDebug("==> last_varKey: " + varA)
        for varB in vars.keys():
             LogDebug("==> varKey: " + varB)
        if last_varKeys == vars.keys():
                LogError("Loop in variables detected.")
        last_varKeys = vars.keys()

        # Attempt to expand and run variable path commands
        LogDebug("Expanding variable path commands...")
        for varpath in unexpandedVarPathCmds.keys():
                varpathitem = unexpandedVarPathCmds[varpath]
                varpathcmd = varpathitem[0]
                ignore = varpathitem[1]
                LogDebug("Expanding variable path command (" + str(varpath) + ") with current value (" + str(varpathcmd) + ").")
                for expanded in vars.keys():
                        varText = "&&" + expanded + "&&"
                        varValue = vars[expanded]
                        varValue = re.sub(r"\\",r"\\\\",varValue)
                        varpathcmd = re.sub(varText,varValue,varpathcmd)
                        LogDebug("Variable Path Command (" + str(varpath) + ") after replacement of (" + str(varText) + ") with (" + str(varValue) + "): " + str(varpathcmd))
                        # update unexpandedVarPathCmds
                        unexpandedVarPathCmds[varpath] = [varpathcmd,ignore]
                # if it's fully expanded now, run the command, update the value, and switch list that it's in
                if isFullyExpanded(varpathcmd):
                        replacementpathvalue=run_command(varpathcmd, ignore)
                        if os_type == "WINDOWS":
                                LogDebug("Updating replacement path for windows: " + str(replacementpathvalue))
                                replacementpathvalue=re.sub(r"\\",r"\\\\",replacementpathvalue)
                        LogDebug("Replacement Path Value: " + str(replacementpathvalue))
                        # update vars array
                        vars[varpath]=replacementpathvalue
                        del unexpandedVarPathCmds[varpath]
                else:
                        LogDebug("Variable Path Command is still not fully expanded.  Saving it to run again.")

        # Attempt to expand and run variable commands
        LogDebug("Expanding variable commands...")
        for var in unexpandedVarCmds.keys():
                varcmditem = unexpandedVarCmds[var]
                varcmd = varcmditem[0]
                ignore = varcmditem[1]
                LogDebug("Expanding variable command (" + str(var) + ") with current value (" + str(varcmd) + ").")
                for expanded in vars.keys():
                        varText = "&&" + expanded + "&&"
                        varValue = vars[expanded]
                        varValue = re.sub(r"\\",r"\\\\",varValue)
                        varcmd = re.sub(varText,varValue,varcmd)
                        LogDebug("Variable Command (" + str(var) + ") after replacement of (" + str(varText) + ") with (" + str(varValue) + "): " + str(varcmd))
                        # update unexpandedVarCmds array
                        unexpandedVarCmds[var]=[varcmd,ignore]
                # if it's fully expanded now, run the command, update the value, and switch list that it's in
                if isFullyExpanded(varcmd):
                        replacementvalue=run_command(varcmd, ignore)
                        LogDebug("Replacement Value: " + str(replacementvalue))
                        vars[var]=replacementvalue
                        del unexpandedVarCmds[var]
                else:
                        LogDebug("Variable Command is still not fully expanded.  Saving it to try again.")

        # Attempt to expand vars
        LogDebug("Expanding variables...")
        for var, value in unexpandedVars.items():
                LogDebug("Expanding variable (" + var + ") with current value (" + value + ").")
                for expanded in vars.keys():
                        varText = "&&" + expanded + "&&"
                        varValue = vars[expanded]
                        varValue = re.sub(r"\\",r"\\\\",varValue)
                        value = re.sub(varText,varValue,value)
                        LogDebug("Variable (" + var + ") after replacement of (" + varText + ") with (" + varValue + "): " + value)
                        # update unexpandedVars array
                        unexpandedVars[var] = value
                # if it's fully expanded now, update the value, and switch list that it's in
                if isFullyExpanded(value):
                        LogDebug("Replacement Value: " + value)
                        vars[var]=value
                        del unexpandedVars[var]
                else:
                        LogDebug("Variable is still not fully expanded.  Saving it to try again.")

        # Attempt to expand varpaths
        LogDebug("Expanding variable paths...")
        for var, path in unexpandedVarPaths.items():
                LogDebug("Expanding variable path (" + var + ") with current value (" + path + ").")
                for expanded in vars.keys():
                        varText = "&&" + expanded + "&&"
                        varValue = vars[expanded]
                        varValue = re.sub(r"\\",r"\\\\",varValue)
                        path = re.sub(varText,varValue,path)
                        LogDebug("Variable (" + var + ") after replacement of (" + varText + ") with (" + varValue + "): " + path)
                        # update unexpandedVarPaths array
                        unexpandedVarPaths[var] = path
                # if it's fully expanded now, update the value, and switch list that it's in
                if isFullyExpanded(path):
                        LogDebug("Replacement Value: " + path)
                        vars[var]=path
                        del unexpandedVarPaths[var]
                else:
                        LogDebug("Variable Path is still not fully expanded.  Saving it to try again.")

        # Define any regular expressions that can be processed
        define_regex_matches()

        # recurse until complete
        expand_variables(last_varKeys)


def isFullyExpanded(value):
        var_expression=".*&&\S+?&&.*"

        # does this variable need expanding?
        var_match = re.match(var_expression,value)
        if var_match == None:
                LogDebug("Value (" + value + ") is completed expanded.")
                return 1
        LogDebug("Value (" + value + ") is not completely expanded.")
        return 0


def select_variables():
        '''
        set all varselect values
        '''
        for var,selectList in varselects.items():
                set = 0
                LogDebug("Setting Variable (" + var + ") from variable list (" + selectList + ").")

                # tokenize selectionList into variables
                variableTokens = selectList.split(",")
                for variableToken in variableTokens:
                        # remove whitespace before/after
                        stripped_variableToken = variableToken.strip()
                        variable_expression = "&&(.*)&&"
                        variable_match = re.match(variable_expression,stripped_variableToken)
                        if variable_match == None:
                               LogWarning("Ignoring improperly formed variable (" + stripped_variableToken + ") in selectList")
                               continue

                        variable_name = variable_match.group(1)

                        try:
                                variable_value = vars[variable_name]
                        except:
                                LogDebug("Ignoring variable name (" + variable_name + ") which was undefined.")
                                continue

                        LogDebug("Variable (" + variable_name + ") has value (" + variable_value + ").")

                        # does the variable have a value?
                        if variable_value == "":
                                LogDebug("Ignoring variable (" + str(variable_match) + ") which has an empty value (" + str(variable_value) + ").")
                                continue
                        else:
                                LogDebug("Setting varible (" + var + ") to value (" + variable_value + ").")
                                vars[var] = variable_value
                                set = 1
                                break
                if not set:
                       # variable was never set, let's set it to an empty string
                       LogWarning("No non-empty selection list value found for variable (" + var + ").  Setting to empty string.")
                       vars[var] = ""


def create_file_config(file, ignoreError):
        '''
        getfile and put in AppConfig object of AppServer
        '''
        LogDebug("Getting file (" + file + ") with ignoreError (" + str(ignoreError) + ").")
        configobject = sensorhelper.newModelObject("com.collation.platform.model.topology.app.AppConfig")
        try:
                configcontent = sensorhelper.getFile(file,os_handle)
                configobject.setContent(configcontent)
                configobject.setParent(app_server)
                result.setConfigContent(configobject)
        except:
                LogDebug("Exception in create_file_config.")
                if ignoreError:
                        LogDebug("Ignoring error as directed.")
                else:
                        LogWarning("Unable to getfile (" + file + ")")


def create_registry_config(key, ignoreError):
        '''
        get registry key and put in AppConfig object of AppServer
        '''
        LogDebug("Getting registry key: " + key)
        try:
                keydata = sensorhelper.queryRegistry(key,os_handle)
                if keydata != "":
                        label = key + ".registry"
                        uri = "config://" + bind_address + "/" + template_name + "/" + label
                        configobject = sensorhelper.newModelObject("com.collation.platform.model.topology.app.AppConfig")
                        configobject.setLabel(str(label))
                        configcontent = sensorhelper.newModelObject("com.collation.platform.model.topology.sys.DataFile")
                        LogDebug("Created 'configcontent'")
                        config_crc = zlib.crc32(keydata)
                        config_size = len(keydata)
                        LogDebug("Registry CONFIG: " + str(config_size) + " CRC: " + str(config_crc))
                        configcontent.setContent(keydata)
                        LogDebug("Completed 'configcontent.setContent'")
                        configcontent.setChecksum(str(config_crc))
                        configcontent.setSize(config_size)
                        configcontent.setURI(uri)
                        LogDebug("Completed configcontent.setURI("+ uri + ")'")
                        configcontent.setFixedPath(label)
                        configobject.setContent(configcontent)
                        LogDebug("Completed 'configobject.setContent'")
                        configobject.setParent(app_server)
                        result.setConfigContent(configobject)
        except:
                LogDebug("Exception in create_registry_config.")
                if ignoreError:
                        LogDebug("Ignoring error as directed.")
                else:
                        LogWarning("Unable to get registry key (" + file + ")")


def complete_path(file):
        original_file = file
        # replace any variables
        exp="(\&\&.+?\&\&)"
        matches = re.findall(exp, file)
        for match in matches:
                LogDebug("Looking for variable (" + str(match) + ") in file (" + str(file) + ").")
                varname = match[2:len(match)-2]
                if vars.has_key(varname):
                        replacementValue = vars[varname]
                        replacementValue = re.sub(r"\\",r"\\\\",replacementValue)
                        LogDebug("Replacing (" + str(match) + ") with (" + str(replacementValue) + ") in starting string (" + str(file) + ").")
                        file = re.sub(match,replacementValue,file)
                        LogDebug("Resulting file (" + str(file) + ").")
                else:
                        LogDebug("No variable (" + str(varname) + ") found.  Ignoring...")
        LogDebug("Original file (" + str(original_file) + ").")
        LogDebug("Complete path (" + str(file) + ").")
        return file


def complete_path_with_update(file):
        original_file = file
        # double the \'s in Windows pathnames
        if os_type == "WINDOWS":
                LogDebug("Updating path for windows for file: " + file)
                file=re.sub(r"\\",r"\\\\",file)

        # replace any variables
        exp="(\&\&.+?\&\&)"
        matches = re.findall(exp, file)
        for match in matches:
                LogDebug("Looking for variable (" + str(match) + ") in file (" + str(file) + ").")
                varname = match[2:len(match)-2]
                if vars.has_key(varname):
                        replacementValue = vars[varname]
                        replacementValue = re.sub(r"\\",r"\\\\",replacementValue)
                        LogDebug("Replacing (" + str(match) + ") with (" + str(replacementValue) + ") in starting string (" + str(file) + ").")
                        file = re.sub(match,replacementValue,file)
                        LogDebug("Resulting file (" + str(file) + ").")
                else:
                        LogDebug("No variable (" + str(varname) + ") found.  Ignoring...")
        LogDebug("Original file (" + str(original_file) + ").")
        LogDebug("Complete path (" + str(file) + ").")
        return file


def create_label(file):
        '''
        Remove the double the \'s in Windows pathnames
        '''
        label=file
        if os_type == "WINDOWS":
                LogDebug("Updating label for windows for file: " + file)
                label=re.sub(r"\\\\",r"\\",file)
        LogDebug("Label: " + label)
        normalizedLabel = normalize_path(label)
        return normalizedLabel


def normalize_path(path):
        '''
        Remove the double the ..'s in pathnames
        '''
        if os_type == "WINDOWS":
                os.sep = "\\"
        else:
                os.sep = "/"
        return os.path.normpath(path)


def get_bind_address():
        try:
                bind_ip = seed.getAddress().getIpAddress().getStringNotation()
        except:
                LogDebug("Could not set bind_ip using from seed address stringNotation.  Trying dotNotation...")
                try:
                        bind_ip = seed.getAddress().getIpAddress().getDotNotation()
                except:
                        LogDebug("Could not set bind_ip from seed address using dotNotation.")
                        bind_ip = '0.0.0.0'

        if bind_ip == '0.0.0.0':
                try:
                        bind_ip = seed.getPrimaryIpAddress().getStringNotation()
                except:
                        LogDebug("Could not set bind_ip from primary IP using stringNotation.  Trying dotNotation...")
                        try:
                                bind_ip = seed.getPrimaryIpAddress().getDotNotation()
                        except:
                                LogDebug("Could not set bind_ip from primary IP using dotNotation.")
        listen_port = '0'
        try:
                listen_port = seed.getAddress().getPortNumber()
        except:
                LogDebug("Could not set listening port.  Defaulting to 0.")
        bind_address = str(bind_ip) + ":" + str(listen_port)
        return bind_address

#---------------------------------------------------------------------------------
# MAIN
#---------------------------------------------------------------------------------
unexpandedVars = {}
unexpandedVarCmds = {}
unexpandedVarPaths = {}
unexpandedVarPathCmds = {}
vars = {}
varselects = {}
varregexs = {}
configFiles = {}
regConfigFiles = {}
commands = {}
configCommands = {}
attrConfigCommands = {}
attrvalues = {}
attrCommands = {}

# The first thing we need to do is get the Objects that are passed to a sensor
(os_handle,result,app_server,seed,log,env) = sensorhelper.init(targets)

# os_handle  -->  Os handle object to target system
# result     -->  Results Object
# app_server -->  AppServer Object that is discovered
# seed       -->  seed object which contains information that was found prior
#                 to this CSX running
# log        -->  Object to write to sensor log
# env        -->  Environment variables available to an AppServer

# Now we can use these objects to add MORE information to the AppServer object
# that has already been found.

template_name = seed.getTemplateName()
LogInfo("***** Starting Template Code *****")

os_type = get_os_type(app_server)
LogInfo("AppServer is on: " + str(os_type))

original_commandline = seed.getProcess().getCmdLine()
process_directory = get_process_directory()
LogInfo("Process directory is: " + process_directory)

environment = seed.getProcess().getEnv()
# environment = app_server.getProcessPools[0].getRuntimeProcesses[0].getEnv()
LogInfo("Process environment is: " + str(environment))

envVariables = {}
parseEnv(environment)

for key, value in envVariables.items():
        LogDebug("--> Variable: " + key)
        LogDebug("--> Value: " + value)

bind_address = get_bind_address()
LogInfo("BindAddress is: " + str(bind_address))

dep_props=Properties()
prop_file=coll_home+"/etc/templates/"+template_name+"_extension.properties"
LogDebug("prop_file is: " + str(prop_file))
try:
        fin = FileInputStream(prop_file)
        dep_props.load(fin)
        fin.close()
except:
        LogDebug("Exception loading properties file.")
        LogError("Could not load Properties from " + str(prop_file))


# initialize configFiles, configCommands, commands, and attrCommands
process_props(dep_props)

# expand any variables within variables
last_varKeys = []
expand_variables(last_varKeys)

# moved to within expand_variables
# create any regex matches on variables
# define_regex_matches()

# replace any selection variables
select_variables()

# for debugging purposes, print out the variables
for var,replacement in vars.items():
        LogDebug("Variable: " + str(var))
        LogDebug(" --> Value: " + str(replacement))

# set the attrvalues
for attr, value in attrvalues.items():
        LogInfo("Setting attr (" + attr + ") to value: " + value)
        try:
                set_attr_to_value(attr, value)
                LogInfo("Done setting attr (" + str(attr) + ") to value: " + str(value))
        except:
                LogDebug("Exception in setting attribute value.")
                LogWarning("Failure setting attr (" + str(attr) + ") to value: " + str(value))


# run the commands
for cmd, ignore in commands.items():
        LogDebug("Command: " + str(cmd))
        full_cmd=complete_path(cmd)
        LogInfo("Running command: " + str(full_cmd))
        run_command(full_cmd, ignore)
        LogInfo("Done running command (" + str(full_cmd) + ").")


# run the attrCommands
for attr in attrCommands.keys():
        attrcmditem = attrCommands[attr]
        attrcmd = attrcmditem[0]
        ignore = attrcmditem[1]
        LogDebug("Attr: " + str(attr))
        full_attrcmd=complete_path(attrcmd)
        LogInfo("Setting attr (" + str(attr) + ") using command: " + str(full_attrcmd))
        set_attr_with_command(attr, full_attrcmd, ignore)
        LogInfo("Done setting attr (" + str(attr) + ") using command: " + str(full_attrcmd))


# run the attrConfigCommands
data = ""
keys = attrConfigCommands.keys()
keys.sort(lambda x,y: cmp(x.lower(),y.lower()))
for key in keys:
        LogDebug("Key: " + key)
        attrcfgcmditem = attrConfigCommands[key]
        attrcfgcmd = attrcfgcmditem[0]
        ignore = attrcfgcmditem[1]
        full_keycmd=complete_path(attrcfgcmd)
        LogInfo("Setting key (" + str(key) + ") using command: " + str(full_keycmd))
        output = run_command(full_keycmd, ignore)
        data = data + key + " = " + output + "\n"
        LogInfo("Done setting key (" + str(key) + ") using command: " + str(full_keycmd))
if data != "":
        create_default_app_config(data)


# run the configCommands and create configContents
for cfgcmd in configCommands.items():
        full_cfgcmd=complete_path(cfgcmd)
        ignore = configCommands(cfgcmd)
        LogInfo("Creating app config file for command: " + str(full_cfgcmd))
        try:
                create_app_config(create_label(full_cfgcmd), full_cfgcmd, ignore)
                LogInfo("Done creating app config for command: " + str(full_cfgcmd))
        except:
                LogDebug("Exception in running configuration command.")
                LogWarning("Failure executing: " + str(full_cfgcmd))


# create the configuration files
for file, ignore in configFiles.items():
        full_file=complete_path_with_update(file)
        LogInfo("Creating config file: " + str(full_file))
        create_file_config(full_file, ignore)
        LogInfo("Done creating config file: " + str(full_file))

# create the registry configuration files
for file, ignore in regConfigFiles.items():
        full_file=complete_path(file)
        LogInfo("Creating registry config file: " + str(full_file))
        try:
                create_registry_config(full_file, ignore)
                LogInfo("Done creating registry config file: " + str(full_file))
        except:
                LogDebug("Exception in creating registry config file.")
                LogWarning("Failure creating registry config file: " + str(full_file))

# call any customizations that may need to be added per template
run_customizations()

LogInfo("***** Leaving Template Code *****")
